Placeholder

Struct Placeholder 

pub struct Placeholder;
Placeholder Parameter type for Parameterized types that is used represent Parameterized::parameter_structure. That is, it is used to replace every nested parameter in a Parameterized type yielding a structure-only representation that can later be used with Parameterized::from_parameters to instantiate a Parameterized value with the same shape but different types of parameters.