pub trait SameAs<T> { }
Helper trait used to encode type equality constraints in the associated type bounds of Parameterized. A type X implements SameAs<Y> only when X and Y are the exact same type.
Parameterized
X
SameAs<Y>
Y